@@ -45,6 +45,7 @@
 #include "parse_opt.h"
 #include "version.h"
 #include "parse_dev.h"
+#include "conffile.h"
 
 #ifndef NFS_PROGRAM
 #define NFS_PROGRAM	(100003)
@@ -80,17 +81,37 @@ struct nfsmount_info {
 				*node,		/* mounted-on dir */
 				*type;		/* "nfs" or "nfs4" */
 	char			*hostname;	/* server's hostname */
+	struct sockaddr_storage	address;	/* server's address */
+	socklen_t		salen;		/* size of server's address */
 
 	struct mount_options	*options;	/* parsed mount options */
 	char			**extra_opts;	/* string for /etc/mtab */
 
+	unsigned long		version;	/* NFS version */
 	int			flags,		/* MS_ flags */
 				fake,		/* actually do the mount? */
 				child;		/* forked bg child? */
-
-	sa_family_t		family;		/* supported address family */
 };
 
+#ifdef MOUNT_CONFIG
+static void nfs_default_version(struct nfsmount_info *mi);
+
+static void nfs_default_version(struct nfsmount_info *mi)
+{
+	extern unsigned long config_default_vers;
+	/*
+	 * Use the default value set in the config file when
+	 * the version has not been explicitly set.
+	 */
+	if (mi->version == 0 && config_default_vers) {
+		if (config_default_vers < 4)
+			mi->version = config_default_vers;
+	}
+}
+#else
+inline void nfs_default_version(struct nfsmount_info *mi) {}
+#endif /* MOUNT_CONFIG */
+
 /*
  * Obtain a retry timeout value based on the value of the "retry=" option.
  *
@@ -132,12 +153,14 @@ static int nfs_append_generic_address_option(const struct sockaddr *sap,
 {
 	char address[NI_MAXHOST];
 	char new_option[512];
+	int len;
 
 	if (!nfs_present_sockaddr(sap, salen, address, sizeof(address)))
 		goto out_err;
 
-	if (snprintf(new_option, sizeof(new_option), "%s=%s",
-					keyword, address) >= sizeof(new_option))
+	len = snprintf(new_option, sizeof(new_option), "%s=%s",
+						keyword, address);
+	if (len < 0 || (size_t)len >= sizeof(new_option))
 		goto out_err;
 
 	if (po_append(options, new_option) != PO_SUCCEEDED)
@@ -198,8 +221,7 @@ static int nfs_append_clientaddr_option(const struct sockaddr *sap,
  * Resolve the 'mounthost=' hostname and append a new option using
  * the resulting address.
  */
-static int nfs_fix_mounthost_option(const sa_family_t family,
-				    struct mount_options *options)
+static int nfs_fix_mounthost_option(struct mount_options *options)
 {
 	struct sockaddr_storage dummy;
 	struct sockaddr *sap = (struct sockaddr *)&dummy;
@@ -210,7 +232,7 @@ static int nfs_fix_mounthost_option(const sa_family_t family,
 	if (!mounthost)
 		return 1;
 
-	if (!nfs_name_to_address(mounthost, family, sap, &salen)) {
+	if (!nfs_name_to_address(mounthost, sap, &salen)) {
 		nfs_error(_("%s: unable to determine mount server's address"),
 				progname);
 		return 0;
@@ -256,57 +278,66 @@ static int nfs_append_sloppy_option(struct mount_options *options)
 	return 1;
 }
 
+static int nfs_set_version(struct nfsmount_info *mi)
+{
+	if (!nfs_nfs_version(mi->options, &mi->version))
+		return 0;
+
+	if (strncmp(mi->type, "nfs4", 4) == 0)
+		mi->version = 4;
+	else {
+		char *option = po_get(mi->options, "proto");
+		if (option && strcmp(option, "rdma") == 0)
+			mi->version = 3;
+	}
+
+	/*
+	 * If we still don't know, check for version-specific
+	 * mount options.
+	 */
+	if (mi->version == 0) {
+		if (po_contains(mi->options, "mounthost") ||
+		    po_contains(mi->options, "mountaddr") ||
+		    po_contains(mi->options, "mountvers") ||
+		    po_contains(mi->options, "mountproto"))
+			mi->version = 3;
+	}
+
+	/*
+	 * If enabled, see if the default version was
+	 * set in the config file
+	 */
+	nfs_default_version(mi);
+	
+	return 1;
+}
+
 /*
- * Set up mandatory NFS mount options.
+ * Set up mandatory non-version specific NFS mount options.
  *
  * Returns 1 if successful; otherwise zero.
  */
 static int nfs_validate_options(struct nfsmount_info *mi)
 {
-	struct sockaddr_storage dummy;
-	struct sockaddr *sap = (struct sockaddr *)&dummy;
-	socklen_t salen = sizeof(dummy);
+	struct sockaddr *sap = (struct sockaddr *)&mi->address;
 
 	if (!nfs_parse_devname(mi->spec, &mi->hostname, NULL))
 		return 0;
 
-	if (!nfs_name_to_address(mi->hostname, mi->family, sap, &salen))
+	mi->salen = sizeof(mi->address);
+	if (!nfs_name_to_address(mi->hostname, sap, &mi->salen))
 		return 0;
 
-	if (strncmp(mi->type, "nfs4", 4) == 0) {
-		if (!nfs_append_clientaddr_option(sap, salen, mi->options))
-			return 0;
-	} else {
-		if (!nfs_fix_mounthost_option(mi->family, mi->options))
-			return 0;
-		if (!mi->fake && !nfs_verify_lock_option(mi->options))
-			return 0;
-	}
+	if (!nfs_set_version(mi))
+		return 0;
 
 	if (!nfs_append_sloppy_option(mi->options))
 		return 0;
 
-	return nfs_append_addr_option(sap, salen, mi->options);
-}
+	if (!nfs_append_addr_option(sap, mi->salen, mi->options))
+		return 0;
 
-/*
- * Distinguish between permanent and temporary errors.
- *
- * Returns 0 if the passed-in error is temporary, thus the
- * mount system call should be retried; returns one if the
- * passed-in error is permanent, thus the mount system call
- * should not be retried.
- */
-static int nfs_is_permanent_error(int error)
-{
-	switch (error) {
-	case ESTALE:
-	case ETIMEDOUT:
-	case ECONNREFUSED:
-		return 0;	/* temporary */
-	default:
-		return 1;	/* permanent */
-	}
+	return 1;
 }
 
 /*
@@ -326,16 +357,14 @@ static int nfs_extract_server_addresses(struct mount_options *options,
 	option = po_get(options, "addr");
 	if (option == NULL)
 		return 0;
-	if (!nfs_string_to_sockaddr(option, strlen(option),
-						nfs_saddr, nfs_salen))
+	if (!nfs_string_to_sockaddr(option, nfs_saddr, nfs_salen))
 		return 0;
 
 	option = po_get(options, "mountaddr");
 	if (option == NULL) {
 		memcpy(mnt_saddr, nfs_saddr, *nfs_salen);
 		*mnt_salen = *nfs_salen;
-	} else if (!nfs_string_to_sockaddr(option, strlen(option),
-						mnt_saddr, mnt_salen))
+	} else if (!nfs_string_to_sockaddr(option, mnt_saddr, mnt_salen))
 		return 0;
 
 	return 1;
@@ -423,76 +452,99 @@ static int nfs_construct_new_options(struct mount_options *options,
  *
  * To handle version and transport protocol fallback properly, we
  * need to parse some of the mount options in order to set up a
- * portmap probe.  Mount options that nfs_rewrite_mount_options()
+ * portmap probe.  Mount options that nfs_rewrite_pmap_mount_options()
  * doesn't recognize are left alone.
  *
- * Returns a new group of mount options if successful; otherwise
- * NULL is returned if some failure occurred.
+ * Returns TRUE if rewriting was successful; otherwise
+ * FALSE is returned if some failure occurred.
  */
-static struct mount_options *nfs_rewrite_mount_options(char *str)
+static int
+nfs_rewrite_pmap_mount_options(struct mount_options *options)
 {
-	struct mount_options *options;
 	struct sockaddr_storage nfs_address;
 	struct sockaddr *nfs_saddr = (struct sockaddr *)&nfs_address;
-	socklen_t nfs_salen;
+	socklen_t nfs_salen = sizeof(nfs_address);
 	struct pmap nfs_pmap;
 	struct sockaddr_storage mnt_address;
 	struct sockaddr *mnt_saddr = (struct sockaddr *)&mnt_address;
-	socklen_t mnt_salen;
+	socklen_t mnt_salen = sizeof(mnt_address);
 	struct pmap mnt_pmap;
+	char *option;
 
-	options = po_split(str);
-	if (!options) {
-		errno = EFAULT;
-		return NULL;
-	}
+	/*
+	 * Skip option negotiation for proto=rdma mounts.
+	 */
+	option = po_get(options, "proto");
+	if (option && strcmp(option, "rdma") == 0)
+		goto out;
 
+	/*
+	 * Extract just the options needed to contact server.
+	 * Bail now if any of these have bad values.
+	 */
 	if (!nfs_extract_server_addresses(options, nfs_saddr, &nfs_salen,
 						mnt_saddr, &mnt_salen)) {
 		errno = EINVAL;
-		goto err;
+		return 0;
+	}
+	if (!nfs_options2pmap(options, &nfs_pmap, &mnt_pmap)) {
+		errno = EINVAL;
+		return 0;
 	}
 
-	nfs_options2pmap(options, &nfs_pmap, &mnt_pmap);
-
-	/* The kernel NFS client doesn't support changing the RPC program
-	 * number for these services, so reset these fields before probing
-	 * the server's ports.  */
+	/*
+	 * The kernel NFS client doesn't support changing the RPC
+	 * program number for these services, so force the value of
+	 * these fields before probing the server's ports.
+	 */
 	nfs_pmap.pm_prog = NFS_PROGRAM;
 	mnt_pmap.pm_prog = MOUNTPROG;
 
+	/*
+	 * If the server's rpcbind service isn't available, we can't
+	 * negotiate.  Bail now if we can't contact it.
+	 */
 	if (!nfs_probe_bothports(mnt_saddr, mnt_salen, &mnt_pmap,
 				 nfs_saddr, nfs_salen, &nfs_pmap)) {
 		errno = ESPIPE;
-		goto err;
+		return 0;
 	}
 
 	if (!nfs_construct_new_options(options, &nfs_pmap, &mnt_pmap)) {
 		errno = EINVAL;
-		goto err;
+		return 0;
 	}
 
+out:
 	errno = 0;
-	return options;
-
-err:
-	po_destroy(options);
-	return NULL;
+	return 1;
 }
 
 /*
  * Do the mount(2) system call.
  *
- * Returns 1 if successful, otherwise zero.
+ * Returns TRUE if successful, otherwise FALSE.
  * "errno" is set to reflect the individual error.
  */
-static int nfs_sys_mount(const struct nfsmount_info *mi, const char *type,
-			 const char *options)
+static int nfs_sys_mount(struct nfsmount_info *mi, struct mount_options *opts)
 {
+	char *options = NULL;
 	int result;
 
-	result = mount(mi->spec, mi->node, type,
-				mi->flags & ~(MS_USER|MS_USERS), options);
+	if (po_join(opts, &options) == PO_FAILED) {
+		errno = EIO;
+		return 0;
+	}
+
+	if (verbose)
+		printf(_("%s: trying text-based options '%s'\n"),
+			progname, options);
+
+	if (mi->fake)
+		return 1;
+
+	result = mount(mi->spec, mi->node, mi->type,
+			mi->flags & ~(MS_USER|MS_USERS), options);
 	if (verbose && result) {
 		int save = errno;
 		nfs_error(_("%s: mount(2): %s"), progname, strerror(save));
@@ -502,127 +554,151 @@ static int nfs_sys_mount(const struct nfsmount_info *mi, const char *type,
 }
 
 /*
- * Retry an NFS mount that failed because the requested service isn't
- * available on the server.
- *
- * Returns 1 if successful.  Otherwise, returns zero.
- * "errno" is set to reflect the individual error.
- *
- * Side effect: If the retry is successful, both 'options' and
- * 'extra_opts' are updated to reflect the mount options that worked.
- * If the retry fails, 'options' and 'extra_opts' are left unchanged.
+ * For "-t nfs vers=2" or "-t nfs vers=3" mounts.
  */
-static int nfs_retry_nfs23mount(struct nfsmount_info *mi)
+static int nfs_try_mount_v3v2(struct nfsmount_info *mi)
 {
-	struct mount_options *retry_options;
-	char *retry_str = NULL;
-	char **extra_opts = mi->extra_opts;
+	struct mount_options *options = po_dup(mi->options);
+	int result = 0;
 
-	retry_options = nfs_rewrite_mount_options(*extra_opts);
-	if (!retry_options)
-		return 0;
-
-	if (po_join(retry_options, &retry_str) == PO_FAILED) {
-		po_destroy(retry_options);
-		errno = EIO;
-		return 0;
+	if (!options) {
+		errno = ENOMEM;
+		return result;
 	}
 
-	if (verbose)
-		printf(_("%s: text-based options (retry): '%s'\n"),
-			progname, retry_str);
-
-	if (!nfs_sys_mount(mi, "nfs", retry_str)) {
-		po_destroy(retry_options);
-		free(retry_str);
-		return 0;
+	if (!nfs_fix_mounthost_option(options)) {
+		errno = EINVAL;
+		goto out_fail;
 	}
-
-	free(*extra_opts);
-	*extra_opts = retry_str;
-	po_replace(mi->options, retry_options);
-	return 1;
-}
-
-/*
- * Attempt an NFSv2/3 mount via a mount(2) system call.  If the kernel
- * claims the requested service isn't supported on the server, probe
- * the server to see what's supported, rewrite the mount options,
- * and retry the request.
- *
- * Returns 1 if successful.  Otherwise, returns zero.
- * "errno" is set to reflect the individual error.
- *
- * Side effect: If the retry is successful, both 'options' and
- * 'extra_opts' are updated to reflect the mount options that worked.
- * If the retry fails, 'options' and 'extra_opts' are left unchanged.
- */
-static int nfs_try_nfs23mount(struct nfsmount_info *mi)
-{
-	char **extra_opts = mi->extra_opts;
-
-	if (po_join(mi->options, extra_opts) == PO_FAILED) {
-		errno = EIO;
-		return 0;
+	if (!mi->fake && !nfs_verify_lock_option(options)) {
+		errno = EINVAL;
+		goto out_fail;
 	}
 
-	if (verbose)
-		printf(_("%s: text-based options: '%s'\n"),
-			progname, *extra_opts);
-
-	if (mi->fake)
-		return 1;
-
-	if (nfs_sys_mount(mi, "nfs", *extra_opts))
-		return 1;
-
 	/*
-	 * The kernel returns EOPNOTSUPP if the RPC bind failed,
-	 * and EPROTONOSUPPORT if the version isn't supported.
+	 * Options we negotiate below may be stale by the time this
+	 * file system is unmounted.  In order to force umount.nfs
+	 * to renegotiate with the server, only write the user-
+	 * specified options, and not negotiated options, to /etc/mtab.
 	 */
-	if (errno != EOPNOTSUPP && errno != EPROTONOSUPPORT)
-		return 0;
-
-	return nfs_retry_nfs23mount(mi);
-}
-
-/*
- * Attempt an NFS v4 mount via a mount(2) system call.
- *
- * Returns 1 if successful.  Otherwise, returns zero.
- * "errno" is set to reflect the individual error.
- */
-static int nfs_try_nfs4mount(struct nfsmount_info *mi)
-{
-	char **extra_opts = mi->extra_opts;
-
-	if (po_join(mi->options, extra_opts) == PO_FAILED) {
-		errno = EIO;
-		return 0;
+	if (po_join(options, mi->extra_opts) == PO_FAILED) {
+		errno = ENOMEM;
+		goto out_fail;
 	}
 
-	if (verbose)
-		printf(_("%s: text-based options: '%s'\n"),
-			progname, *extra_opts);
+	if (!nfs_rewrite_pmap_mount_options(options))
+		goto out_fail;
 
-	if (mi->fake)
-		return 1;
+	result = nfs_sys_mount(mi, options);
 
-	return nfs_sys_mount(mi, "nfs4", *extra_opts);
+out_fail:
+	po_destroy(options);
+	return result;
 }
 
 /*
- * Perform either an NFSv2/3 mount, or an NFSv4 mount system call.
+ * For "-t nfs -o vers=4" or "-t nfs4" mounts.
+ */
+static int nfs_try_mount_v4(struct nfsmount_info *mi)
+{
+	struct sockaddr *sap = (struct sockaddr *)&mi->address;
+	struct mount_options *options = po_dup(mi->options);
+	int result = 0;
+
+	if (!options) {
+		errno = ENOMEM;
+		return result;
+	}
+
+	if (mi->version == 0) {
+		if (po_append(options, "vers=4") == PO_FAILED) {
+			errno = EINVAL;
+			goto out_fail;
+		}
+	}
+
+	if (!nfs_append_clientaddr_option(sap, mi->salen, options)) {
+		errno = EINVAL;
+		goto out_fail;
+	}
+
+	/*
+	 * Update option string to be recorded in /etc/mtab.
+	 */
+	if (po_join(options, mi->extra_opts) == PO_FAILED) {
+		errno = ENOMEM;
+		goto out_fail;
+	}
+
+	result = nfs_sys_mount(mi, options);
+
+out_fail:
+	po_destroy(options);
+	return result;
+}
+
+/*
+ * This is a single pass through the fg/bg loop.
  *
- * Returns 1 if successful.  Otherwise, returns zero.
+ * Returns TRUE if successful, otherwise FALSE.
  * "errno" is set to reflect the individual error.
  */
 static int nfs_try_mount(struct nfsmount_info *mi)
 {
-	if (strncmp(mi->type, "nfs4", 4) == 0)
-		return nfs_try_nfs4mount(mi);
-	else
-		return nfs_try_nfs23mount(mi);
+	int result = 0;
+
+	switch (mi->version) {
+	case 0:
+		if (linux_version_code() > MAKE_VERSION(2, 6, 31)) {
+			errno = 0;
+			result = nfs_try_mount_v4(mi);
+			if (errno != EPROTONOSUPPORT) {
+				/* 
+				 * To deal with legacy Linux servers that don't
+				 * automatically export a pseudo root, retry
+				 * ENOENT errors using version 3
+				 */
+				if (errno != ENOENT)
+					break;
+			}
+		}
+	case 2:
+	case 3:
+		result = nfs_try_mount_v3v2(mi);
+		break;
+	case 4:
+		result = nfs_try_mount_v4(mi);
+		break;
+	default:
+		errno = EIO;
+	}
+
+	return result;
+}
+
+/*
+ * Distinguish between permanent and temporary errors.
+ *
+ * Basically, we retry if communication with the server has
+ * failed so far, but fail immediately if there is a local
+ * error (like a bad mount option).
+ *
+ * ESTALE is also a temporary error because some servers
+ * return ESTALE when a share is temporarily offline.
+ *
+ * Returns 1 if we should fail immediately, or 0 if we
+ * should retry.
+ */
+static int nfs_is_permanent_error(int error)
+{
+	switch (error) {
+	case ESTALE:
+	case ETIMEDOUT:
+	case ECONNREFUSED:
+		return 0;	/* temporary */
+	default:
+		return 1;	/* permanent */
+	}
 }
 
 /*
@@ -785,11 +861,6 @@ int nfsmount_string(const char *spec, const char *node, const char *type,
 		.flags		= flags,
 		.fake		= fake,
 		.child		= child,
-#ifdef IPV6_SUPPORTED
-		.family		= AF_UNSPEC,	/* either IPv4 or v6 */
-#else
-		.family		= AF_INET,	/* only IPv4 */
-#endif
 	};
 	int retval = EX_FAIL;
 
